#!/usr/bin/env python

#This is the parray class (pitched array) that serves as
#a complement to pycuda.GPUArray.
#The intention here is to automatically create 2D or 3D
#array with a pitched structure
#and varies operation on it.

import numpy as np
import pycuda.driver as cuda
from pytools import memoize

from . import parray_utils as pu
from past.builtins import long

""" utilities"""
@memoize
def _splay_backend(n, M):
    block_count = 6 * cuda.Context.get_device().MULTIPROCESSOR_COUNT
    if M <= 1:
        block = (256, 1, 1)
    else:
        block = (32, 8, 1)

    return (block_count, 1), block


def splay(n, M):
    return _splay_backend(n, M)


def _get_common_dtype(obj1, obj2):
    """Return the 'least common multiple' of dtype of obj1 and obj2."""
    return (obj1.dtype.type(0) + obj2.dtype.type(0)).dtype


def _get_inplace_dtype(obj1, obj2):
    """
    Returns the dtype of obj1,
    Raise error if
    1) obj1 is real and obj2 is complex
    2) obj1 is integer and obj2 is floating

    Parameters
    ----------
    obj1 : numpy.ndarray like array
    obj2 : numpy.ndarray like array

    Returns
    -------
    out : np.dtype
    """
    if isrealobj(obj1):
        if iscomplexobj(obj2):
            raise TypeError("Cannot cast complex dtype to real dtype")
    if issubclass(obj1.dtype.type, np.integer):
        if issubclass(obj2.dtype.type, (np.floating, np.complexfloating)):
            raise TypeError("Cannot cast floating to integer")
    return obj1.dtype

def _get_common_dtype_with_scalar(scalar, obj1):
    """
    return the common dtype between a native scalar (int, float, complex)
    and the dtype of an ndarray like array.

    Parameters
    ----------
    scalar : { int, float, complex }
    obj1 : numpy.ndarray like array.

    """
    if issubclass(type(scalar), (int, float, np.integer, np.floating)):
        return obj1.dtype
    elif issubclass(type(scalar), (complex, np.complexfloating)):
        if isrealobj(obj1):
            return floattocomplex(obj1.dtype)
        else:
            return obj1.dtype
    else:
        raise TypeError("scalar type is not supported")


def _get_inplace_dtype_with_scalar(scalar, obj1):
    """
    Returns the dtype of obj1,
    Raise error if
    1) obj1 is real and obj2 is complex
    2) obj1 is integer and obj2 is floating

    Parameters
    ----------
    obj1 : numpy.ndarray like array
    obj2 : numpy.ndarray like array

    Returns
    -------
    out : np.dtype

    """
    if isrealobj(obj1):
        if issubclass(type(scalar), (complex, np.complexfloating)):
            raise TypeError("Cannot cast complex dtype to real dtype")
    if issubclass(obj1.dtype.type, np.integer):
        if issubclass(
                type(scalar),
                (float, complex, np.floating, np.complexfloating)):
            raise TypeError("Cannot cast floating to integer")
    return obj1.dtype


def _pd(shape):
    """ Returns the product of all element of shape except shape[0]. """
    s = 1
    for dim in shape[1:]:
        s *= dim
    return s


def _assignshape(shape, axis, value):
    a = []
    for i in range(len(shape)):
        if i == axis:
            a.append(value)
        else:
            a.append(shape[i])
    return tuple(a)


def PitchTrans(shape, dst, dst_ld, src, src_ld, dtype, aligned=False, _async = False, stream = None):
    """
    Wrapper around cuda.Memcpy2D
    Enable pitched memory transfer.

    Parameters
    ----------
    shape : tuple of ints
        shape of the 2D array to be transferred.
    dst : { cuda.DeviceAllocation, int, long }
        pointer to the device memory to be transferred to.
    dst_ld: int
        leading dimension (pitch) of destination.
    src : { pycuda.driver.DeviceAllocation, int, long }
        pointer to the device memory to be transferred from.
    src_ld : int
        leading dimension (pitch) of source.

    Optional Parameters
    -------------------
    aligned : bool
        if aligned is False, tolerate device-side misalignment for
        device-to-device copies that may lead to loss of copy bandwidth.
        (default: False).
    async : bool
        use asynchronous transfer (default: False).
    stream : pycuda.driver.stream
        specify the cuda stream (default: None).

    """
    size = np.dtype(dtype).itemsize

    trans = cuda.Memcpy2D()
    trans.src_pitch = src_ld * size
    if isinstance(src, (cuda.DeviceAllocation, int, long)):
        trans.set_src_device(src)
    else:
        trans.set_src_host(src)

    trans.dst_pitch = dst_ld * size
    if isinstance(dst, (cuda.DeviceAllocation, int, long)):
        trans.set_dst_device(dst)
    else:
        trans.set_dst_host(dst)

    trans.width_in_bytes = _pd(shape) * size
    trans.height = int(shape[0])

    if _async:
        trans(stream)
    else:
        trans(aligned = aligned)

"""end of utilities"""

class PitchArray(object):
    def __init__(self, shape, dtype, gpudata=None, pitch = None, base = None):
        """
        Create a PitchArray

        Parameters
        ----------
        shape: tuple of ints
            shape of the array
        dtype: np.dtype
            dtype of the array
        gpudata: pycuda.driver.DeviceAllocation
            DeviceAllocation object indicating the device memory allocated
        pitch: int
            if gpudata is specified and pitch is True,
            gpudata will be treated as if it was allocated by
            cudaMallocPitch with pitch
        base: PitchArray
            base PitchArray

        Attributes:
        -----------
        .shape: shape of self
        .size:  number of elements of the array
        .mem_size: number of elements of total memory allocated
        .ld: leading dimension
        .M: 1 if self is a vector, shape[0] otherwise
        .N: self.size if self is a vector
            product of shape[1] and shape[2] otherwise
        .gpudata: DeviceAllocation
        .ndim: number of dimensions
        .dtype: dtype of array
        .nbytes: total memory allocated for the array in bytes
        .base: base PitchArray

        Note:
        -----
        1. any 1-dim shape will result in a row vector with
        new shape as (1, shape) operations of PitchArray
        is elementwise operation

        2. only support array of dimension up to 3.

        """

        try:
            tmpshape = []
            s = 1
            for dim in shape:
                dim = int(dim)
                assert isinstance(dim, int)
                s *= dim
                tmpshape.append(dim)

            self.shape = tuple(tmpshape)
        except TypeError:
            s = int(shape)
            assert isinstance(s, int)
            if s:
                self.shape = (1, s)
            else:
                self.shape = (0, 0)

        if len(self.shape) == 1:
            self.shape = (1, self.shape[0])

        self.ndim = len(self.shape)

        if self.ndim > 3:
            raise ValueError("Only support array of dimension leq 3")

        self.dtype = np.dtype(dtype)
        self.size = s

        if gpudata is None:
            if self.size:
                if _pd(self.shape) == 1 or self.shape[0] == 1:
                    self.gpudata = cuda.mem_alloc(
                                   self.size * self.dtype.itemsize)
                    self.mem_size = self.size
                    self.ld = _pd(self.shape)
                    self.M = 1
                    self.N = self.size
                else:
                    self.gpudata, pitch = cuda.mem_alloc_pitch(
                        int(_pd(self.shape) * np.dtype(dtype).itemsize),
                        self.shape[0], np.dtype(dtype).itemsize)
                    self.ld = pitch // np.dtype(dtype).itemsize
                    self.mem_size = self.ld * self.shape[0]
                    self.M = self.shape[0]
                    self.N = _pd(self.shape)
            else:
                self.gpudata = None
                self.M = 0
                self.N = 0
                self.ld = 0
                self.mem_size = 0
            self.base = base
        else:
            # assumed that the device memory was also allocated
            # by mem_alloc_pitch is required by the shape
            # Not performed to allow base
            # assert gpudata.__class__ == cuda.DeviceAllocation

            if self.size:
                self.gpudata = gpudata
                if _pd(self.shape) == 1 or self.shape[0] == 1:
                    self.mem_size = self.size
                    self.ld = _pd(self.shape)
                    self.M = 1
                    self.N = self.size
                else:
                    if pitch is None:
                        pitch = int(np.ceil(
                                float(_pd(self.shape)
                                * np.dtype(dtype).itemsize)
                                // 512) * 512)
                    else:
                        assert pitch == int(np.ceil(
                                        float(_pd(self.shape)
                                        * np.dtype(dtype).itemsize)
                                        // 512) * 512)

                    self.ld = pitch // np.dtype(dtype).itemsize
                    self.mem_size = self.ld * self.shape[0]
                    self.M = self.shape[0]
                    self.N = _pd(self.shape)
            else:
                self.gpudata = None
                self.M = 0
                self.N = 0
                self.ld = 0
                self.mem_size = 0
                print("warning: shape may not be assigned properly")
            self.base = base
        self.nbytes = self.dtype.itemsize * self.mem_size
        self._grid, self._block = splay(self.mem_size, self.M)

    def set(self, ary):
        """
        Set PitchArray with an numpy.ndarray

        Parameter:
        ----------
        ary: num.ndarray
             must have the same shape as self if ndim > 2,
             and same length as self if ndim == 1

        """
        assert ary.ndim <= 3
        assert ary.dtype == ary.dtype
        assert ary.size == self.size

        if self.size:
            if self.M == 1:
                cuda.memcpy_htod(int(self.gpudata), ary)
            else:
                PitchTrans(self.shape, int(self.gpudata),
                           self.ld, ary, _pd(self.shape), self.dtype)

    def set_async(self, ary, stream=None):
        """
        Set PitchArray with an numpy.ndarray
        using asynchrous memroy transfer

        Parameter:
        ----------
        ary: num.ndarray pagelocked
             must have the same shape as self if ndim > 2,
             and same length as self if ndim == 1
             must be created by cuda.HostAllocation

        Optional Parameter:
        -------------------
        stream : pycuda.driver.Stream

        """
        assert ary.ndim <= 3
        assert ary.dtype == ary.dtype
        assert ary.size == self.size

        if ary.base.__class__ != cuda.HostAllocation:
            raise TypeError("asynchronous memory trasfer "
                            "requires pagelocked numpy array")

        if self.size:
            if self.M == 1:
                cuda.memcpy_htod_async(int(self.gpudata), ary, stream)
            else:
                PitchTrans(self.shape, int(self.gpudata), self.ld, ary,
                           _pd(self.shape), self.dtype, _async = True,
                           stream = stream)

    def get(self, ary = None, pagelocked = False):
        """
        Get the PitchArray to an ndarray

        Optional Parameters:
        --------------------
        ary: numpy.ndarray
            if specified, will transfer device memory to ary's memory
            if None, create a new array
            (default: None).
        pagelocked: bool
            if True, create a pagelocked ndarray
            if False, create a regular ndarray
            (default: False)

        """
        if ary is None:
            if pagelocked:
                ary = cuda.pagelocked_empty(self.shape, self.dtype)
            else:
                ary = np.empty(self.shape, self.dtype)
        else:
            assert ary.size == self.size
            assert ary.dtype == ary.dtype


        if self.size:
            if self.M == 1:
                cuda.memcpy_dtoh(ary, int(self.gpudata))
            else:
                PitchTrans(self.shape, ary, _pd(self.shape),
                           int(self.gpudata), self.ld, self.dtype)

        return ary


    def get_async(self, stream = None, ary = None):
        """
        Get the PitchArray to an ndarray asynchronously

        Optional Parameters:
        --------------------
        stream: pycuda.driver.Stream
            (default: None)
        ary: numpy.ndarray
            if specified, will transfer device memory to ary's memory,
            must be pagelocked
            if None, create a new array
            (default: None)

        """
        if ary is None:
            ary = cuda.pagelocked_empty(self.shape, self.dtype)
        else:
            assert ary.size == self.size
            assert ary.dtype == ary.dtype
            if ary.base.__class__ != cuda.HostAllocation:
                raise TypeError("asynchronous memory trasfer "
                                "requires pagelocked numpy array")

        if self.size:
            if self.M == 1:
                cuda.memcpy_dtoh_async(ary, int(self.gpudata), stream)
            else:
                PitchTrans(self.shape, ary, _pd(self.shape),
                           int(self.gpudata), self.ld, self.dtype,
                           _async = True, stream = stream)

        return ary

    def __str__(self):
        return str(self.get())

    def __repr__(self):
        return repr(self.get())

    def __hash__(self):
        raise TypeError("PitchArrays are not hashable.")

    def _new_like_me(self, dtype = None):
        if dtype is None:
            dtype = self.dtype
        return self.__class__(self.shape, dtype)

    """""""""
    Operators:
    operators defined by __op__(self, other) returns new PitchArray
    operators defined by op(self, other)  perform inplace operation
        if inplace cannot be done, error raises
    operators defined by __iop__(self, other) also perform inplace
        operation if possbile, otherwise returns a new PitchArray
    """""""""

    def __add__(self, other):
        """
        Addition

        Parameters
        ----------
        other: scalar or Pitcharray

        Returns
        -------
        out : PitchArray
            A new PitchArray
        """
        if isinstance(other, PitchArray):
            if self.shape != other.shape:
                raise ValueError("array dimension misaligned")
            result = self._new_like_me(_get_common_dtype(self, other))
            if self.size:
                if self.M == 1:
                    func = pu.get_addarray_function(
                        self.dtype, other.dtype, result.dtype, pitch = False)
                    func.prepared_call(
                        self._grid, self._block, result.gpudata,
                        self.gpudata, other.gpudata, self.size)
                else:
                    func = pu.get_addarray_function(
                        self.dtype, other.dtype, result.dtype, pitch = True)
                    func.prepared_call(
                        self._grid, self._block, self.M, self.N,
                        result.gpudata, result.ld, self.gpudata,
                        self.ld, other.gpudata, other.ld)
            return result
        elif issubclass(type(other), (float, int, complex, np.integer,
                                      np.floating, np.complexfloating)):
            dtype = _get_common_dtype_with_scalar(other, self)
            if other == 0:
                return self.astype(dtype)
            else:
                result = self._new_like_me(dtype)
                if self.size:
                    if self.M == 1:
                        func = pu.get_addscalar_function(
                            self.dtype, dtype, pitch = False)
                        func.prepared_call(
                            self._grid, self._block, result.gpudata,
                            self.gpudata, other, self.size)
                    else:
                        func = pu.get_addscalar_function(
                            self.dtype, dtype, pitch = True)
                        func.prepared_call(
                            self._grid, self._block, self.M, self.N,
                            result.gpudata, result.ld, self.gpudata,
                            self.ld, other)
                return result
        else:
            raise TypeError("type of object to be added is not supported")

    __radd__ = __add__

    def __sub__(self, other):
        """
        Substraction
        self - other

        Parameters
        ----------
        other : scalar or Pitcharray

        Returns
        -------
        out : PitchArray
            A new PitchArray
        """
        if isinstance(other, PitchArray):
            if self.shape != other.shape:
                raise ValueError("array dimension misaligned")
            result = self._new_like_me(_get_common_dtype(self, other))
            if self.size:
                if self.M == 1:
                    func = pu.get_subarray_function(
                        self.dtype, other.dtype, result.dtype, pitch = False)
                    func.prepared_call(
                        self._grid, self._block, result.gpudata,
                        self.gpudata, other.gpudata, self.size)
                else:
                    func = pu.get_subarray_function(
                        self.dtype, other.dtype, result.dtype, pitch = True)
                    func.prepared_call(
                        self._grid, self._block, self.M, self.N,
                        result.gpudata, result.ld, self.gpudata,
                        self.ld, other.gpudata, other.ld)
            return result
        elif issubclass(type(other), (float, int, complex, np.integer,
                                      np.floating, np.complexfloating)):
            dtype = _get_common_dtype_with_scalar(other, self)
            if other == 0:
                return self.astype(dtype)
            else:
                result = self._new_like_me(dtype)
                if self.size:
                    if self.M == 1:
                        func = pu.get_subscalar_function(
                            self.dtype, dtype, pitch = False)
                        func.prepared_call(
                            self._grid, self._block, result.gpudata,
                            self.gpudata, other, self.size)
                    else:
                        func = pu.get_subscalar_function(
                            self.dtype, dtype, pitch = True)
                        func.prepared_call(
                            self._grid, self._block, self.M, self.N,
                            result.gpudata, result.ld, self.gpudata,
                            self.ld, other)
                return result
        else:
            raise TypeError("type of object to be substracted "
                            "is not supported")

    def __rsub__(self, other):
        """
        Being substracted
        Other - self

        Parameters
        ----------
        other: scalar or Pitcharray

        Returns
        -------
        out : PitchArray
            A new PitchArray
        """
        """
        # this part is not neccessary
        if isinstance(other, PitchArray):
            if self.shape != other.shape:
                raise ValueError("array dimension misaligned")
            result = self._new_like_me(_get_common_dtype(self, other))
            if self.size:
                if self.M == 1:
                    func = pu.get_subarray_function(
                        other.dtype, self.dtype, result.dtype, pitch = False)
                    func.prepared_call(
                        self._grid, self._block, result.gpudata,
                        other.gpudata, self.gpudata, self.size)
                else:
                    func = pu.get_subarray_function(
                        other.dtype, self.dtype, result.dtype, pitch = True)
                    func.prepared_call(
                        self._grid, self._block, self.M, self.N,
                        result.gpudata, result.ld, other.gpudata,
                        other.ld, self.gpudata, self.ld)
            return result
        """
        if issubclass(type(other), (float, int, complex, np.integer,
                                      np.floating, np.complexfloating)):
            dtype = _get_common_dtype_with_scalar(other, self)
            result = self._new_like_me(dtype)
            if self.size:
                if self.M == 1:
                    func = pu.get_scalarsub_function(
                        self.dtype, dtype, pitch = False)
                    func.prepared_call(
                        self._grid, self._block, result.gpudata,
                        self.gpudata, other, self.size)
                else:
                    func = pu.get_scalarsub_function(
                        self.dtype, dtype, pitch = True)
                    func.prepared_call(
                        self._grid, self._block, self.M, self.N,
                        result.gpudata, result.ld, self.gpudata,
                        self.ld, other)
            return result
        else:
            raise TypeError("type of object to substract from "
                            "is not supported")

    def __mul__(self, other):
        """
        Multiply

        Parameters
        ----------
        other: scalar or Pitcharray

        Returns
        -------
        out : PitchArray
            A new PitchArray
        """
        if isinstance(other, PitchArray):
            if self.shape != other.shape:
                raise ValueError("array dimension misaligned")
            result = self._new_like_me(_get_common_dtype(self, other))
            if self.size:
                if self.M == 1:
                    func = pu.get_mularray_function(
                        self.dtype, other.dtype, result.dtype, pitch = False)
                    func.prepared_call(
                        self._grid, self._block, result.gpudata,
                        self.gpudata, other.gpudata, self.size)
                else:
                    func = pu.get_mularray_function(
                        self.dtype, other.dtype, result.dtype, pitch = True)
                    func.prepared_call(
                        self._grid, self._block, self.M, self.N,
                        result.gpudata, result.ld, self.gpudata,
                        self.ld, other.gpudata, other.ld)
            return result
        elif issubclass(type(other), (float, int, complex, np.integer,
                                      np.floating, np.complexfloating)):
            dtype = _get_common_dtype_with_scalar(other, self)
            if other == 1.0:
                return self.astype(dtype)
            else:
                result = self._new_like_me(dtype)
                if self.size:
                    if self.M == 1:
                        func = pu.get_mulscalar_function(
                            self.dtype, dtype, pitch = False)
                        func.prepared_call(
                            self._grid, self._block, result.gpudata,
                            self.gpudata, other, self.size)
                    else:
                        func = pu.get_mulscalar_function(
                            self.dtype, dtype, pitch = True)
                        func.prepared_call(
                            self._grid, self._block, self.M, self.N,
                            result.gpudata, result.ld, self.gpudata,
                            self.ld, other)
                return result
        else:
            raise TypeError("type of object to be multiplied "
                            "is not supported")

    __rmul__ = __mul__

    def __div__(self, other):
        """
        Division
        self / other

        Parameters
        ----------
        other: scalar or Pitcharray

        Returns
        -------
        out : PitchArray
            A new PitchArray
        """
        if isinstance(other, PitchArray):
            if self.shape != other.shape:
                raise ValueError("array dimension misaligned")
            result = self._new_like_me(_get_common_dtype(self, other))
            if self.size:
                if self.M == 1:
                    func = pu.get_divarray_function(
                        self.dtype, other.dtype, result.dtype, pitch = False)
                    func.prepared_call(
                        self._grid, self._block, result.gpudata,
                        self.gpudata, other.gpudata, self.size)
                else:
                    func = pu.get_divarray_function(
                        self.dtype, other.dtype, result.dtype, pitch = True)
                    func.prepared_call(
                        self._grid, self._block, self.M, self.N,
                        result.gpudata, result.ld, self.gpudata,
                        self.ld, other.gpudata, other.ld)
            return result
        elif issubclass(type(other), (float, int, complex, np.integer,
                                      np.floating, np.complexfloating)):
            dtype = _get_common_dtype_with_scalar(other, self)
            if other == 1.0:
                return self.astype(dtype)
            else:
                result = self._new_like_me(dtype)
                if self.size:
                    if self.M == 1:
                        func = pu.get_divscalar_function(
                            self.dtype, dtype, pitch = False)
                        func.prepared_call(
                            self._grid, self._block, result.gpudata,
                            self.gpudata, other, self.size)
                    else:
                        func = pu.get_divscalar_function(
                            self.dtype, dtype, pitch = True)
                        func.prepared_call(
                            self._grid, self._block, self.M, self.N,
                            result.gpudata, result.ld, self.gpudata,
                            self.ld, other)
                return result
        else:
            raise TypeError("type of object to be divided "
                            "is not supported")

    def __rdiv__(self, other):
        """
        Being divided
        other / self

        Parameters
        ----------
        other: scalar or Pitcharray

        Returns
        -------
        out : PitchArray
            A new PitchArray

        """
        """
        # this part it not necessary
        if isinstance(other, PitchArray):
            if self.shape != other.shape:
                raise ValueError("array dimension misaligned")
            result = self._new_like_me(_get_common_dtype(self, other))
            if self.size:
                if self.M == 1:
                    func = pu.get_divarray_function(
                        other.dtype, self.dtype, result.dtype, pitch = False)
                    func.prepared_call(
                        self._grid, self._block, result.gpudata,
                        other.gpudata, self.gpudata, self.size)
                else:
                    func = pu.get_divarray_function(
                        other.dtype, self.dtype, result.dtype, pitch = True)
                    func.prepared_call(
                        self._grid, self._block, self.M, self.N,
                        result.gpudata, result.ld, other.gpudata,
                        other.ld, self.gpudata, self.ld)
            return result
        """
        if issubclass(type(other), (float, int, complex, np.integer,
                                      np.floating, np.complexfloating)):
            dtype = _get_common_dtype_with_scalar(other, self)
            result = self._new_like_me(dtype)
            if self.size:
                if self.M == 1:
                    func = pu.get_scalardiv_function(
                        self.dtype, dtype, pitch = False)
                    func.prepared_call(
                        self._grid, self._block, result.gpudata,
                        self.gpudata, other, self.size)
                else:
                    func = pu.get_scalardiv_function(
                        self.dtype, dtype, pitch = True)
                    func.prepared_call(
                        self._grid, self._block, self.M, self.N,
                        result.gpudata, result.ld, self.gpudata,
                        self.ld, other)
            return result
        else:
            raise TypeError("type of object to be divided from"
                            "is not supported")

    def __neg__(self):
        """
        Take negative value
        """
        return 0-self

    def __iadd__(self, other):
        """
        add to self inplace

        Parameters
        ----------
        other: scalar or Pitcharray

        Returns
        -------
        out : PitchArray (self)

        Note
        ----
        If other is complex, self is required to be a complex
        """
        if isinstance(other, PitchArray):
            if self.shape != other.shape:
                raise ValueError("array dimension misaligned")
            dtype = _get_common_dtype(self, other)
            if self.dtype == dtype:
                result = self
            else:
                result = self._new_like_me(dtype = dtype)

            if self.size:
                if self.M == 1:
                    func = pu.get_addarray_function(
                        self.dtype, other.dtype, result.dtype, pitch = False)
                    func.prepared_call(
                        self._grid, self._block, result.gpudata,
                        self.gpudata, other.gpudata, self.size)
                else:
                    func = pu.get_addarray_function(
                        self.dtype, other.dtype, result.dtype, pitch = True)
                    func.prepared_call(self._grid, self._block,
                        self.M, self.N, result.gpudata, result.ld,
                        self.gpudata, self.ld, other.gpudata, other.ld)
            return result
        elif issubclass(type(other), (float, int, complex, np.integer,
                                      np.floating, np.complexfloating)):
            dtype = _get_common_dtype_with_scalar(other, self)
            if other == 0:
                return self
            else:
                if self.dtype != dtype:
                    result = self._new_like_me(dtype)
                else:
                    result = self
                if self.size:
                    if self.M == 1:
                        func = pu.get_addscalar_function(
                            self.dtype, dtype, pitch = False)
                        func.prepared_call(
                            self._grid, self._block, result.gpudata,
                            self.gpudata, other, self.size)
                    else:
                        func = pu.get_addscalar_function(
                            self.dtype, dtype, pitch = True)
                        func.prepared_call(
                            self._grid, self._block, self.M, self.N,
                            result.gpudata, self.ld, self.gpudata,
                            self.ld, other)
                return result
        else:
            raise TypeError("type of object to be added"
                            "is not supported")

    def __isub__(self, other):
        """
        Substracted other inplace

        Parameters
        ----------
        other: scalar or Pitcharray

        Returns
        -------
        out : PitchArray (self)
        """
        if isinstance(other, PitchArray):
            if self.shape != other.shape:
                raise ValueError("array dimension misaligned")
            dtype = _get_common_dtype(self, other)
            if self.dtype == dtype:
                result = self
            else:
                result = self._new_like_me(dtype = dtype)

            if self.size:
                if self.M == 1:
                    func = pu.get_subarray_function(
                        self.dtype, other.dtype, result.dtype, pitch = False)
                    func.prepared_call(
                        self._grid, self._block, result.gpudata,
                        self.gpudata, other.gpudata, self.size)
                else:
                    func = pu.get_subarray_function(
                        self.dtype, other.dtype, result.dtype, pitch = True)
                    func.prepared_call(
                        self._grid, self._block, self.M, self.N,
                        result.gpudata, result.ld, self.gpudata,
                        self.ld, other.gpudata, other.ld)
            return result
        elif issubclass(type(other), (float, int, complex, np.integer,
                                      np.floating, np.complexfloating)):
            dtype = _get_common_dtype_with_scalar(other, self)
            if other == 0:
                return self
            else:
                if self.dtype != dtype:
                    result = self._new_like_me(dtype)
                else:
                    result = self
                if self.size:
                    if self.M == 1:
                        func = pu.get_subscalar_function(
                            self.dtype, dtype, pitch = False)
                        func.prepared_call(
                            self._grid, self._block, result.gpudata,
                            self.gpudata, other, self.size)
                    else:
                        func = pu.get_subscalar_function(
                            self.dtype, dtype, pitch = True)
                        func.prepared_call(
                            self._grid, self._block, self.M, self.N,
                            result.gpudata, self.ld, self.gpudata,
                            self.ld, other)
                return result
        else:
            raise TypeError("type of object to be substracted"
                            "is not supported")

    def __imul__(self, other):
        """
        Multiplied by other
        inplace if possible

        Parameters
        ----------
        other: scalar or Pitcharray

        Returns
        -------
        out : PitchArray (self)
        """
        if isinstance(other, PitchArray):
            if self.shape != other.shape:
                raise ValueError("array dimension misaligned")
            dtype = _get_common_dtype(self, other)
            if self.dtype == dtype:
                result = self
            else:
                result = self._new_like_me(dtype = dtype)

            if self.size:
                if self.M == 1:
                    func = pu.get_mularray_function(
                        self.dtype, other.dtype, result.dtype, pitch = False)
                    func.prepared_call(
                        self._grid, self._block, result.gpudata,
                        self.gpudata, other.gpudata, self.size)
                else:
                    func = pu.get_mularray_function(
                        self.dtype, other.dtype, result.dtype, pitch = True)
                    func.prepared_call(
                        self._grid, self._block, self.M, self.N,
                        result.gpudata, result.ld, self.gpudata,
                        self.ld, other.gpudata, other.ld)
            return result
        elif issubclass(type(other), (float, int, complex, np.integer,
                                      np.floating, np.complexfloating)):
            dtype = _get_common_dtype_with_scalar(other, self)
            if other == 1.0:
                return self
            else:
                if self.dtype != dtype:
                    result = self._new_like_me(dtype)
                else:
                    result = self
                if self.size:
                    if self.M == 1:
                        func = pu.get_mulscalar_function(
                            self.dtype, dtype, pitch = False)
                        func.prepared_call(
                            self._grid, self._block, result.gpudata,
                            self.gpudata, other, self.size)
                    else:
                        func = pu.get_mulscalar_function(
                            self.dtype, dtype, pitch = True)
                        func.prepared_call(
                            self._grid, self._block, self.M, self.N,
                            result.gpudata, self.ld, self.gpudata,
                            self.ld, other)
                return result
        else:
            raise TypeError("type of object to be multiplied"
                            "is not supported")

    def __idiv__(self, other):
        """
        Divided by other
        inplace if possible

        Parameters
        ----------
        other: scalar or Pitcharray

        Returns
        -------
        out : PitchArray (self)
        """
        if isinstance(other, PitchArray):
            if self.shape != other.shape:
                raise ValueError("array dimension misaligned")
            dtype = _get_common_dtype(self, other)
            if self.dtype == dtype:
                result = self
            else:
                result = self._new_like_me(dtype = dtype)

            if self.size:
                if self.M == 1:
                    func = pu.get_divarray_function(
                        self.dtype, other.dtype, result.dtype, pitch = False)
                    func.prepared_call(
                        self._grid, self._block, result.gpudata,
                        self.gpudata, other.gpudata, self.size)
                else:
                    func = pu.get_divarray_function(
                        self.dtype, other.dtype, result.dtype, pitch = True)
                    func.prepared_call(
                        self._grid, self._block, self.M, self.N,
                        result.gpudata, result.ld, self.gpudata,
                        self.ld, other.gpudata, other.ld)
            return result
        elif issubclass(type(other), (float, int, complex, np.integer,
                                      np.floating, np.complexfloating)):
            dtype = _get_common_dtype_with_scalar(other, self)
            if other == 1.0:
                return self
            else:
                if self.dtype != dtype:
                    result = self._new_like_me(dtype)
                else:
                    result = self
                if self.size:
                    if self.M == 1:
                        func = pu.get_divscalar_function(
                            self.dtype, dtype, pitch = False)
                        func.prepared_call(
                            self._grid, self._block, result.gpudata,
                            self.gpudata, other, self.size)
                    else:
                        func = pu.get_divscalar_function(
                            self.dtype, dtype, pitch = True)
                        func.prepared_call(
                            self._grid, self._block, self.M, self.N,
                            result.gpudata, self.ld, self.gpudata,
                            self.ld, other)
                return result
        else:
            raise TypeError("type of object to be divided"
                            "is not supported")

    def __pow__(self, other):
        if isinstance(other, PitchArray):
            if self.shape != other.shape:
                raise ValueError("array dimension misaligned")
            result = self._new_like_me(_get_common_dtype(self, other))
            if self.size:
                if self.M == 1:
                    func = pu.get_powarray_function(
                        self.dtype, other.dtype, result.dtype, pitch = False)
                    func.prepared_call(
                        self._grid, self._block, result.gpudata,
                        self.gpudata, other.gpudata, self.size)
                else:
                    func = pu.get_powarray_function(
                        self.dtype, other.dtype, result.dtype, pitch = True)
                    func.prepared_call(
                        self._grid, self._block, self.M, self.N,
                        result.gpudata, result.ld, self.gpudata,
                        self.ld, other.gpudata, other.ld)
            return result
        elif issubclass(type(other), (float, int, complex, np.integer,
                                      np.floating, np.complexfloating)):
            dtype = _get_common_dtype_with_scalar(other, self)
            if other == 0:
                return self.astype(dtype)
            else:
                result = self._new_like_me(dtype)
                if self.size:
                    if self.M == 1:
                        func = pu.get_powscalar_function(
                            self.dtype, dtype, pitch = False)
                        func.prepared_call(
                            self._grid, self._block, result.gpudata,
                            self.gpudata, other, self.size)
                    else:
                        func = pu.get_powscalar_function(
                            self.dtype, dtype, pitch = True)
                        func.prepared_call(
                            self._grid, self._block, self.M, self.N,
                            result.gpudata, result.ld, self.gpudata,
                            self.ld, other)
                return result
        else:
            raise TypeError("type of object to be powered is not supported")

    def __ipow__(self, other):
        """
        add to self inplace

        Parameters
        ----------
        other: scalar or Pitcharray

        Returns
        -------
        out : PitchArray (self)

        Note
        ----
        If other is complex, self is required to be a complex
        """
        if isinstance(other, PitchArray):
            if self.shape != other.shape:
                raise ValueError("array dimension misaligned")
            dtype = _get_common_dtype(self, other)
            if self.dtype == dtype:
                result = self
            else:
                result = self._new_like_me(dtype = dtype)

            if self.size:
                if self.M == 1:
                    func = pu.get_powarray_function(
                        self.dtype, other.dtype, result.dtype, pitch = False)
                    func.prepared_call(
                        self._grid, self._block, result.gpudata,
                        self.gpudata, other.gpudata, self.size)
                else:
                    func = pu.get_powarray_function(
                        self.dtype, other.dtype, result.dtype, pitch = True)
                    func.prepared_call(self._grid, self._block,
                        self.M, self.N, result.gpudata, result.ld,
                        self.gpudata, self.ld, other.gpudata, other.ld)
            return result
        elif issubclass(type(other), (float, int, complex, np.integer,
                                      np.floating, np.complexfloating)):
            dtype = _get_common_dtype_with_scalar(other, self)
            if other == 0:
                return self.astype(dtype)
            else:
                if self.dtype != dtype:
                    result = self._new_like_me(dtype)
                else:
                    result = self
                if self.size:
                    if self.M == 1:
                        func = pu.get_powscalar_function(
                            self.dtype, dtype, pitch = False)
                        func.prepared_call(
                            self._grid, self._block, result.gpudata,
                            self.gpudata, other, self.size)
                    else:
                        func = pu.get_powscalar_function(
                            self.dtype, dtype, pitch = True)
                        func.prepared_call(
                            self._grid, self._block, self.M, self.N,
                            result.gpudata, self.ld, self.gpudata,
                            self.ld, other)
                return result
        else:
            raise TypeError("type of object to be powered"
                            "is not supported")


    def add(self, other):
        """
        add other to self
        inplace
        """
        if isinstance(other, PitchArray):
            if self.shape != other.shape:
                raise ValueError("array dimension misaligned")
            dtype = _get_inplace_dtype(self, other)
            if self.size:
                if self.M == 1:
                    func = pu.get_addarray_function(
                        self.dtype, other.dtype, self.dtype, pitch = False)
                    func.prepared_call(
                        self._grid, self._block, self.gpudata,
                        self.gpudata, other.gpudata, self.size)
                else:
                    func = pu.get_addarray_function(
                        self.dtype, other.dtype, self.dtype, pitch = True)
                    func.prepared_call(self._grid, self._block,
                        self.M, self.N, self.gpudata, self.ld,
                        self.gpudata, self.ld, other.gpudata, other.ld)
            return self
        elif issubclass(type(other), (float, int, complex, np.integer,
                                      np.floating, np.complexfloating)):
            dtype = _get_inplace_dtype_with_scalar(other, self)
            if other != 0:
                if self.size:
                    if self.M == 1:
                        func = pu.get_addscalar_function(
                            self.dtype, self.dtype, pitch = False)
                        func.prepared_call(
                            self._grid, self._block, self.gpudata,
                            self.gpudata, other, self.size)
                    else:
                        func = pu.get_addscalar_function(
                            self.dtype, self.dtype, pitch = True)
                        func.prepared_call(
                            self._grid, self._block, self.M, self.N,
                            self.gpudata, self.ld, self.gpudata,
                            self.ld, other)
            return self
        else:
            raise TypeError("type of object to be added"
                            "is not supported")

    def sub(self, other):
        """
        substract other from self
        inplace
        """
        if isinstance(other, PitchArray):
            if self.shape != other.shape:
                raise ValueError("array dimension misaligned")
            dtype = _get_inplace_dtype(self, other)
            if self.size:
                if self.M == 1:
                    func = pu.get_subarray_function(
                        self.dtype, other.dtype, self.dtype, pitch = False)
                    func.prepared_call(
                        self._grid, self._block, self.gpudata,
                        self.gpudata, other.gpudata, self.size)
                else:
                    func = pu.get_subarray_function(
                        self.dtype, other.dtype, self.dtype, pitch = True)
                    func.prepared_call(self._grid, self._block,
                        self.M, self.N, self.gpudata, self.ld,
                        self.gpudata, self.ld, other.gpudata, other.ld)
            return self
        elif issubclass(type(other), (float, int, complex, np.integer,
                                      np.floating, np.complexfloating)):
            dtype = _get_inplace_dtype_with_scalar(other, self)
            if other != 0:
                if self.size:
                    if self.M == 1:
                        func = pu.get_subscalar_function(
                            self.dtype, self.dtype, pitch = False)
                        func.prepared_call(
                            self._grid, self._block, self.gpudata,
                            self.gpudata, other, self.size)
                    else:
                        func = pu.get_subscalar_function(
                            self.dtype, self.dtype, pitch = True)
                        func.prepared_call(
                            self._grid, self._block, self.M, self.N,
                            self.gpudata, self.ld, self.gpudata,
                            self.ld, other)
            return self
        else:
            raise TypeError("type of object to be substracted"
                            "is not supported")

    def rsub(self, other):
        """
        substract other by self
        inplace
        """
        if isinstance(other, PitchArray):
            if self.shape != other.shape:
                raise ValueError("array dimension misaligned")
            dtype = _get_inplace_dtype(self, other)
            if self.size:
                if self.M == 1:
                    func = pu.get_subarray_function(
                        other.dtype, self.dtype, self.dtype, pitch = False)
                    func.prepared_call(
                        self._grid, self._block, self.gpudata,
                        other.gpudata, self.gpudata, self.size)
                else:
                    func = pu.get_subarray_function(
                        other.dtype, self.dtype, self.dtype, pitch = True)
                    func.prepared_call(
                        self._grid, self._block, self.M, self.N,
                        self.gpudata, self.ld, other.gpudata,
                        other.ld, self.gpudata, self.ld)
            return self
        elif issubclass(type(other), (float, int, complex, np.integer,
                                      np.floating, np.complexfloating)):
            dtype = _get_inplace_dtype_with_scalar(other, self)
            if self.size:
                if self.M == 1:
                    func = pu.get_scalarsub_function(
                        self.dtype, self.dtype, pitch = False)
                    func.prepared_call(
                        self._grid, self._block, self.gpudata,
                        self.gpudata, other, self.size)
                else:
                    func = pu.get_scalarsub_function(
                        self.dtype, self.dtype, pitch = True)
                    func.prepared_call(
                        self._grid, self._block, self.M, self.N,
                        self.gpudata, self.ld, self.gpudata,
                        self.ld, other)
            return self
        else:
            raise TypeError("type of object to substract from"
                            "is not supported")

    def mul(self, other):
        """
        multiply other with self
        inplace
        """
        if isinstance(other, PitchArray):
            if self.shape != other.shape:
                raise ValueError("array dimension misaligned")
            dtype = _get_inplace_dtype(self, other)
            if self.size:
                if self.M == 1:
                    func = pu.get_mularray_function(
                        self.dtype, other.dtype, self.dtype, pitch = False)
                    func.prepared_call(
                        self._grid, self._block, self.gpudata,
                        self.gpudata, other.gpudata, self.size)
                else:
                    func = pu.get_mularray_function(
                        self.dtype, other.dtype, self.dtype, pitch = True)
                    func.prepared_call(self._grid, self._block,
                        self.M, self.N, self.gpudata, self.ld,
                        self.gpudata, self.ld, other.gpudata, other.ld)
            return self
        elif issubclass(type(other), (float, int, complex, np.integer,
                                      np.floating, np.complexfloating)):
            dtype = _get_inplace_dtype_with_scalar(other, self)
            if other != 1:
                if self.size:
                    if self.M == 1:
                        func = pu.get_mulscalar_function(
                            self.dtype, self.dtype, pitch = False)
                        func.prepared_call(
                            self._grid, self._block, self.gpudata,
                            self.gpudata, other, self.size)
                    else:
                        func = pu.get_mulscalar_function(
                            self.dtype, self.dtype, pitch = True)
                        func.prepared_call(
                            self._grid, self._block, self.M, self.N,
                            self.gpudata, self.ld, self.gpudata,
                            self.ld, other)
            return self
        else:
            raise TypeError("type of object to be multiplied"
                            "is not supported")

    def div(self, other):
        """
        divide other from self
        inplace
        """
        if isinstance(other, PitchArray):
            if self.shape != other.shape:
                raise ValueError("array dimension misaligned")
            dtype = _get_inplace_dtype(self, other)
            if self.size:
                if self.M == 1:
                    func = pu.get_divarray_function(
                        self.dtype, other.dtype, self.dtype, pitch = False)
                    func.prepared_call(
                        self._grid, self._block, self.gpudata,
                        self.gpudata, other.gpudata, self.size)
                else:
                    func = pu.get_divarray_function(
                        self.dtype, other.dtype, self.dtype, pitch = True)
                    func.prepared_call(self._grid, self._block,
                        self.M, self.N, self.gpudata, self.ld,
                        self.gpudata, self.ld, other.gpudata, other.ld)
            return self
        elif issubclass(type(other), (float, int, complex, np.integer,
                                      np.floating, np.complexfloating)):
            dtype = _get_inplace_dtype_with_scalar(other, self)
            if other != 1:
                if self.size:
                    if self.M == 1:
                        func = pu.get_divscalar_function(
                            self.dtype, self.dtype, pitch = False)
                        func.prepared_call(
                            self._grid, self._block, self.gpudata,
                            self.gpudata, other, self.size)
                    else:
                        func = pu.get_divscalar_function(
                            self.dtype, self.dtype, pitch = True)
                        func.prepared_call(
                            self._grid, self._block, self.M, self.N,
                            self.gpudata, self.ld, self.gpudata,
                            self.ld, other)
            return self
        else:
            raise TypeError("type of object to be divided"
                            "is not supported")

    def rdiv(self, other):
        """
        divide other by self
        inplace
        """
        if isinstance(other, PitchArray):
            if self.shape != other.shape:
                raise ValueError("array dimension misaligned")
            dtype = _get_inplace_dtype(self, other)
            if self.size:
                if self.M == 1:
                    func = pu.get_divarray_function(
                        other.dtype, self.dtype, self.dtype, pitch = False)
                    func.prepared_call(
                        self._grid, self._block, self.gpudata,
                        other.gpudata, self.gpudata, self.size)
                else:
                    func = pu.get_divarray_function(
                        other.dtype, self.dtype, self.dtype, pitch = True)
                    func.prepared_call(
                        self._grid, self._block, self.M, self.N,
                        self.gpudata, self.ld, other.gpudata,
                        other.ld, self.gpudata, self.ld)
            return self
        elif issubclass(type(other), (float, int, complex, np.integer,
                                      np.floating, np.complexfloating)):
            dtype = _get_inplace_dtype_with_scalar(other, self)
            if self.size:
                if self.M == 1:
                    func = pu.get_scalardiv_function(
                        self.dtype, self.dtype, pitch = False)
                    func.prepared_call(
                        self._grid, self._block, self.gpudata,
                        self.gpudata, other, self.size)
                else:
                    func = pu.get_scalardiv_function(
                        self.dtype, dtype, pitch = True)
                    func.prepared_call(
                        self._grid, self._block, self.M, self.N,
                        self.gpudata, self.ld, self.gpudata,
                        self.ld, other)
            return self
        else:
            raise TypeError("type of object to divide from"
                            "is not supported")
    def neg(self):
        """
        Take the negative of self inplace

        Returns
        -------
        self
        """
        if self.size:
            if self.M == 1:
                func = pu.get_scalarsub_function(
                    self.dtype, self.dtype, pitch = False)
                func.prepared_call(
                    self._grid, self._block, self.gpudata,
                    self.gpudata, 0, self.size)
            else:
                func = pu.get_scalarsub_function(
                    self.dtype, self.dtype, pitch = True)
                func.prepared_call(
                    self._grid, self._block, self.M, self.N,
                    self.gpudata, self.ld, self.gpudata,
                    self.ld, 0)
        return self

    def fill(self, value, stream=None):
        """
        Fill all entries of self with value

        Parameters:
        ----------------------------------
        value: scalar
               Value to be filled
        stream: pycuda.driver.stream
                for asynchronous execution
        """
        if self.size:
            if self.M == 1:
                func = pu.get_fill_function(self.dtype, pitch = False)
                func.prepared_call(
                    self._grid, self._block, self.size, self.gpudata, value)
            else:
                func = pu.get_fill_function(
                    self.dtype, pitch = True)

                func.prepared_call(
                    self._grid, self._block, self.M, self.N,
                    self.gpudata, self.ld, self.dtype.type(value))

    def copy(self, result=None):
        """
        Returns a duplicated copy of self
        """
        if not result:
            result = self._new_like_me()
        else:
            assert(self.dtype == result.dtype)
            assert(self.mem_size == result.mem_size)
        if self.size:
            cuda.memcpy_dtod(
                result.gpudata, self.gpudata,
                self.mem_size * self.dtype.itemsize)
        return result

    def real(self):
        """
        Returns the real part of self
        """
        if self.dtype == np.complex128:
            result = self._new_like_me(dtype = np.float64)
        elif self.dtype == np.complex64:
            result = self._new_like_me(dtype = np.float32)
        else:
            return self

        if self.size:
            if self.M == 1:
                func = pu.get_realimag_function(
                    self.dtype, real = True, pitch = False)
                func.prepared_call(
                    self._grid, self._block, result.gpudata,
                    self.gpudata, self.size)
            else:
                func = pu.get_realimag_function(
                    self.dtype, real = True, pitch = True)
                func.prepared_call(
                    self._grid, self._block, self.M, self.N,
                    result.gpudata, result.ld, self.gpudata, self.ld)
        return result

    def imag(self):
        """
        returns the imaginary part of self
        """
        if self.dtype == np.complex128:
            result = self._new_like_me(dtype = np.float64)
        elif self.dtype == np.complex64:
            result = self._new_like_me(dtype = np.float32)
        else:
            return zeros_like(self)

        if self.size:
            if self.M == 1:
                func = pu.get_realimag_function(
                    self.dtype, real = False, pitch = False)
                func.prepared_call(
                    self._grid, self._block, result.gpudata,
                    self.gpudata, self.size)
            else:
                func = pu.get_realimag_function(
                    self.dtype, real = False, pitch = True)
                func.prepared_call(
                    self._grid, self._block, self.M, self.N,
                    result.gpudata, result.ld, self.gpudata, self.ld)
        return result

    def abs(self):
        """
        returns the absolute value of self
        """
        if self.dtype in [np.complex128, np.float64]:
            result = self._new_like_me(dtype = np.float64)
        elif self.dtype in [np.complex64, np.float32]:
            result = self._new_like_me(dtype = np.float32)
        else:
            result = self._new_like_me()

        if self.M == 1:
            func = pu.get_abs_function(self.dtype, pitch = False)
            func.prepared_call(
                self._grid, self._block, result.gpudata,
                self.gpudata, self.size)
        else:
            func = pu.get_abs_function(
                self.dtype, pitch = True)
            func.prepared_call(
                self._grid, self._block, self.M, self.N,
                result.gpudata, result.ld, self.gpudata, self.ld)
        return result

    def conj(self, inplace = True):
        """
        returns the conjuation of self.

        Paramters:
        ----------
        inplace: bool (optional)
            if inplace is True, conjugation will be performed in place
            (default: True)

        """
        if self.dtype in [np.complex64, np.complex128]:
            if inplace:
                result = self
            else:
                result = self._new_like_me()

            if self.M == 1:
                func = pu.get_conj_function(self.dtype, pitch = False)
                func.prepared_call(
                    self._grid, self._block, result.gpudata,
                    self.gpudata, self.size)
            else:
                func = pu.get_conj_function(
                    self.dtype, pitch = True)
                func.prepared_call(
                    self._grid, self._block, self.M, self.N,
                    result.gpudata, result.ld, self.gpudata, self.ld)
        else:
            result = self
        return result

    def reshape(self, shape, inplace = True):
        """
        reshape the shape of self to "shape"

        Paramters:
        ----------
        shape : tuple of ints
            the new shape to be reshaped to
        inplace : bool (optional)
            if True, enforce to keep the device memory of self if possible.
            If the above is not possible, e.g. the new shape requires larger
            memory size, or if inplace is False, return a new PitchArray.
            (default:  True)
        """

        sx = 1
        for dim in self.shape:
            sx *= dim

        s = 1
        flag = False
        n = 0
        axis = 0
        idx = -1
        for dim in shape:
            if dim == -1:
                flag = True
                n += 1
                idx = axis
            else:
                s *= dim
            axis += 1

        if flag:
            if n > 1:
                raise ValueError("can only specify one unknown dimension")
            else:
                if sx % s == 0:
                    shape = _assignshape(shape, idx, int(sx // s))
                else:
                    raise ValueError("cannot infer the size "
                                     "of the remaining axis")
        else:
            if s != sx:
                raise ValueError("total size of new array must be unchanged")

        if inplace:
            if shape[0] == self.shape[0]:
                #self.shape = shape
                return PitchArray(shape = shape,
                                  dtype = self.dtype,
                                  gpudata = int(self.gpudata),
                                  base = self)
            else:# The case when self is a vector
                if self.M == 1 and any(b == 1 for b in shape):
                    return PitchArray(shape = shape,
                                      dtype = self.dtype,
                                      gpudata = int(self.gpudata),
                                      base = self)
                else:
                    raise ValueError("cannot resize inplacely")


        result = PitchArray(shape, self.dtype)
        func = pu.get_resize_function(self.dtype)
        #func.set_block_shape(256,1,1)
        func.prepared_call(
            self._grid, (256,1,1), self.shape[0], _pd(self.shape),
            result.shape[0], _pd(result.shape), result.gpudata,
            result.ld, self.gpudata, self.ld)
        return result

    def astype(self, dtype):
        """
        Convert dtype of self to dtype

        Parameters:
        -----------
        dtype: np.dtype
               dtype of the returned array

        """
        dtype = np.dtype(dtype)
        if self.dtype == dtype:
            return self.copy()
        else:
            result = self._new_like_me(dtype = dtype)

            if self.size:
                if self.M == 1:
                    func = pu.get_astype_function(
                        dtype, self.dtype, pitch = False)
                    func.prepared_call(
                        self._grid, self._block, result.gpudata,
                        self.gpudata, self.size)
                else:
                    func = pu.get_astype_function(
                        dtype, self.dtype, pitch = True)
                    func.prepared_call(
                        self._grid, self._block, self.M, self.N,
                        result.gpudata, result.ld, self.gpudata, self.ld)
            return result

    def T(self):
        """
        Returns the transpose
        PitchArray must be 2 dimensional
        """

        if len(self.shape) > 2:
            raise ValueError("transpose only apply to 2D matrix")

        shape_t = self.shape[::-1]

        if self.M == 1:
            result = self.copy()
            result.shape = shape_t
            result.ld = _pd(result.shape)
            return result

        result = PitchArray(shape_t, self.dtype)
        if self.size:
            func = pu.get_transpose_function(self.dtype)
            func.prepared_call(self._grid, self._block, self.shape[0],
                               self.shape[1], result.gpudata,
                               result.ld, self.gpudata, self.ld)
        return result

    def H(self):
        """
        Returns the conjugate transpose
        PitchArray must be 2 dimensional
        """
        if len(self.shape) > 2:
            raise ValueError("transpose only apply to 2D matrix")

        shape_t = self.shape[::-1]
        if self.M == 1:
            result = conj(self)
            result.shape = shape_t
            result.ld = _pd(result.shape)
            return result

        result = PitchArray(shape_t, self.dtype)

        if self.size:
            func = pu.get_transpose_function(self.dtype, conj = True)
            func.prepared_call(self._grid, self._block, self.shape[0],
                               self.shape[1], result.gpudata, result.ld,
                               self.gpudata, self.ld)
        return result

    def copy_rows(self, start, stop, step = 1):
        """
        Extract rows of self to form a new array.

        Parameters:
        -----------
        start: int
            first row to be extracted
        stop: int
            last row to be extracted
        step: int (optional, default: 1)
            extract every step row.

        """
        nrows = len(range(start,stop,step))
        if nrows:

            if self.ndim == 2:
                shape = (nrows, self.shape[1])
            else:
                shape = (nrows, self.shape[1], self.shape[2])
        else:
            if self.ndim == 2:
                shape = (nrows, 0)
            else:
                shape = (nrows, 0, 0)

        result = PitchArray(shape, self.dtype)

        if nrows > 1:
            PitchTrans(
                shape, result.gpudata, result.ld,
                int(self.gpudata) + start*self.ld*self.dtype.itemsize,
                self.ld * step, self.dtype)
        elif nrows == 1:
            cuda.memcpy_dtod(
                result.gpudata,
                int(int(self.gpudata) + start*self.ld*self.dtype.itemsize),
                self.dtype.itemsize * _pd(shape))
        return result

    def view(self, dtype = None):
        """
        New view of array with the same data (similar to numpy.ndarary.view)

        Optional Parameters:
        --------------------
        dtype : numpy.dtype
            Data-type descriptor of the returned view

        """

        if dtype is None:
            dtype = self.dtype
        old_itemsize = self.dtype.itemsize
        itemsize = np.dtype(dtype).itemsize

        if self.shape[-1] * old_itemsize % itemsize != 0:
            raise ValueError("new type not compatible with array")

        shape = self.shape[:-1] + (self.shape[-1] * old_itemsize // itemsize,)

        return PitchArray(shape=shape, dtype=dtype, gpudata=int(self.gpudata),
                          base=self)

    def __getitem__(self, idx):
        """
        only support slicing a chunk of consecutive rows
        """
        if idx == ():
            return self

        if isinstance(idx, slice):
            start = idx.start
            stop = idx.stop
            step = idx.step

            if start is None:
                start = 0
            if stop is None:
                stop = self.shape[0]
            if step is None:
                step = 1

            if step != 1:
                raise NotImplementedError("non-consecutive slicing is not "
                                          "implemented yet")
        elif isinstance(idx, tuple):
            if len(idx) > 1:
                axis = 1;
                for tmp in idx[1:]:
                    start = tmp.start
                    stop = tmp.stop
                    step = tmp.step
                    if start is None:
                        start = 0
                    if stop is None:
                        stop = self.shape[axis]
                    if step is None:
                        step = 1

                    if (start != 0 or stop != self.shape[axis] or
                        step != 1):
                        if self.M != 1:
                            raise NotImplementedError(
                            "slicing only supported on axis = 0")
                    axis += 1

                start = idx[0].start
                stop = idx[0].stop
                step = idx[0].step
                if start is None and stop is None and step is None:
                    if self.M == 1:
                        if len(idx) == 2:
                            start = idx[1].start
                            stop = idx[1].stop
                            step = idx[1].step
                if start is None:
                    start = 0
                if stop is None:
                    stop = self.shape[0]
                if step is None:
                    step = 1

                if step != 1:
                    raise NotImplementedError("non-consecutive slicing is "
                                              "not implemented yet")
        elif isinstance(idx, int):
            start = idx
            stop = idx + 1
            step = 1
        else:
            raise ValueError("non-slice indexing not supported: %s" % (idx,))

        maxshape = self.size if self.M == 1 else self.shape[0]
        if stop > maxshape:
            stop = maxshape
            from warnings import warn
            warn("array slicing larger than array size, "
                 "reduce to allowed size")

        if self.M == 1:
            if self.shape[0] == 1:
                shape = (1,stop-start)
            else:
                shape = (stop-start,1)
            return PitchArray(shape=shape, dtype = self.dtype,
                              gpudata = int(int(self.gpudata) +
                                    start*self.dtype.itemsize),
                              base = self)
        else:
            if self.ndim == 2:
                return PitchArray(shape=(stop-start,self.shape[1]),
                                  dtype = self.dtype,
                                  gpudata = int(int(self.gpudata) +
                                        start*self.dtype.itemsize*self.ld),
                                  base = self)
            else:
                return PitchArray(shape=(stop-start,self.shape[1],
                                          self.shape[2]),
                                  dtype = self.dtype,
                                  gpudata = int(int(self.gpudata) +
                                        start*self.dtype.itemsize*self.ld),
                                  base = self)


def to_gpu(ary):
    """
    Transfer a numpy ndarray to a PitchArray
    """
    result = PitchArray(ary.shape, ary.dtype)
    result.set(ary)
    return result


def to_gpu_async(ary, stream = None):
    """
    Transfer a pagelocked numpy ndarray to a PitchArray
    """
    result = PitchArray(ary.shape, ary.dtype)
    result.set_async(ary, stream)


empty = PitchArray


def empty_like(other_ary):
    """
    Create an empty PitchArray, whose shape
    and dtype is the same as other_ary
    """
    result = PitchArray(other_ary.shape, other_ary.dtype)
    return result


def zeros(shape, dtype):
    """
    Create a PitchArray with all entry equal 0

    Parameter:
    -------------------------------
    shape: tuple or list of int
           Shape of the new array
    dtype: np.dtype
           dtype of the new array
    """
    result = PitchArray(shape, dtype)
    result.fill(0)
    return result

def zeros_like(other_ary):
    """
    Create a PitchArray with all entry equal 0, whose shape
    and dtype is the same as other_ary
    """
    result = PitchArray(other_ary.shape, other_ary.dtype)
    result.fill(0)
    return result


def ones(shape, dtype):
    """
    Create a PitchArray with all entry equal 1

    Parameter:
    -------------------------------
    shape: tuple or list of int
           Shape of the new array
    dtype: np.dtype
           dtype of the new array
    """
    result = PitchArray(shape, dtype)
    result.fill(1)
    return result

def ones_like(other_ary):
    """
    Create a PitchArray with all entry equal 1, whose shape
    and dtype is the same as other_ary
    """
    result = PitchArray(other_ary.shape, other_ary.dtype)
    result.fill(1)
    return result


def make_pitcharray(dptr, shape, dtype, linear = False, pitch=None):
    """
    Create a PitchArray from a DeviceAllocation pointer

    Parameters:
    -----------
    dptr : pycuda.driver.DeviceAllocation
        pointer to the device memory
    shape : tuple of ints
        shape of the array
    dtype : numpy.dtype
        data type of the array
    linear : bool (optional, default: False)
        "True" indicates the device memory is a linearly allocated
        "False" indicates the device memory is allocated by
        cudaMallocPitch, and pitch must be provided
    pitch : int
        pitch of the array

    """
    if linear:
        result = PitchArray(shape, dtype)
        if result.size:
            if result.M == 1:
                cuda.memcpy_dtod(
                    result.gpudata, dptr, result.mem_size * dtype.itemsize)
            else:
                PitchTrans(
                    shape, result.gpudata, result.ld, dptr, _pd(shape), dtype)
    else:
        result = PitchArray(shape, dtype, gpudata=dptr, pitch = pitch)
    return result


def arrayg2p(other_gpuarray):
    """
    Convert a GPUArray to a PitchArray

    Parameter:
    ---------------------------------
    other_gpuarray : pycuda.GPUArray

    """
    result = make_pitcharray(
        other_gpuarray.gpudata, other_gpuarray.shape,
        other_gpuarray.dtype, linear = True)
    return result


def arrayp2g(pary):
    """
    Convert a PitchArray to a GPUArray

    Parameter:
    ----------
    pary: PitchArray

    """
    from pycuda.gpuarray import GPUArray
    result = GPUArray(pary.shape, pary.dtype)
    if pary.size:
        if pary.M == 1:
            cuda.memcpy_dtod(
                result.gpudata, pary.gpudata,
                pary.mem_size * pary.dtype.itemsize)
        else:
            PitchTrans(pary.shape, result.gpudata, _pd(result.shape),
                       pary.gpudata, pary.ld, pary.dtype)
    return result


def conj(pary):
    """
    Returns the conjugation of 2D PitchArray
    same as PitchArray.conj(), but create a new copy

    Parameter:
    ----------
    pary: PitchArray

    """
    return pary.conj(inplace = False)


def reshape(pary, shape):
    """
    Returns reshaped of 2D PitchArray
    same as PitchArray.reshape(), but always create a new copy

    Parameter:
    ----------
    pary: PitchArray

    """
    return pary.reshape(shape, inplace = False)


def iscomplexobj(pary):
    """ Check if the array dtype is complex """
    return issubclass(pary.dtype.type, np.complexfloating)


def isrealobj(pary):
    """ Check if the array dtype is real """
    return not iscomplexobj(pary)


def issingle(dtype):
    """ Check if dtype is single floating point """
    if dtype in [np.float32, np.complex64]:
        return True
    elif dtype in [np.float64, np.complex128]:
        return False
    else:
        raise TypeError("input dtype " + str(dtype) +
                        "not understood")


def floattocomplex(dtype):
    """ Conver dtype from real to corresponding complex dtype """
    dtype = dtype.type if isinstance(dtype, np.dtype) else dtype
    if issubclass(dtype, np.complexfloating):
        outdtype = dtype
    elif dtype == np.float32:
        outdtype = np.complex64
    elif dtype == np.float64:
        outdtype = np.complex128
    else:
        raise TypeError("input dtype " + str(dtype) +
                        " cannot be translated to complex floating")
    return np.dtype(outdtype)


def complextofloat(dtype):
    """ convert dtype from complex to corresponding real dtype """
    dtype = dtype.type if isinstance(dtype, np.dtype) else dtype
    if not issubclass(dtype, np.complexfloating):
        outdtype = dtype
    elif dtype == np.complex64:
        outdtype = np.float32
    elif dtype == np.complex128:
        outdtype = np.float64
    else:
        raise TypeError("input dtype " + str(dtype) +
                        " cannot be translated to floating")
    return np.dtype(outdtype)

def make_complex(real, imag):
    """
    Create a complex array using two real arrays

    Parameters
    ---------------------------------
    real: PitchArray
          The real part
    imag: PitchArray
          The imaginary part
          shape of real and imag must be the same

    Returns
    ---------------------------------
    out: PitchArray
         The complex array

    """
    if isinstance(real, np.ndarray):
        real = to_gpu(real)
    if isinstance(imag, np.ndarray):
        imag = to_gpu(imag)

    if real.shape != imag.shape:
        raise ValueError("real and imaginary parts must have the same shape")
    if iscomplexobj(real) or iscomplexobj(imag):
        raise TypeError("real and imaginary parts must be real array")
    dtype = _get_common_dtype(real, imag)
    if dtype in [np.int32, np.float32]:
        dtype = np.dtype(np.complex64)
    elif dtype in [np.int64, np.float64]:
        dtype = np.dtype(np.complex128)
    else:
        dtype = np.dtype(np.complex64)

    result = empty(real.shape, dtype)
    if result.size:
        if result.M == 1:
            func = pu.get_complex_function(
                real.dtype, imag.dtype, dtype, pitch = False)
            func.prepared_call(
                result._grid, result._block, result.gpudata,
                real.gpudata, imag.gpudata, result.size)
        else:
            func = pu.get_complex_function(
                real.dtype, imag.dtype, dtype)
            func.prepared_call(
                result._grid, result._block, result.M, result.N,
                result.gpudata, result.ld, real.gpudata, imag.gpudata,
                real.ld)
    return result

def angle(array):
    """ Returns the angle of each element in a complex array """
    if isrealobj(array):
        if issingle(array.dtype):
            return parray.zeros(array.shape, np.float32)
        else:
            return parray.zeros(array.shape, np.double)
    else:
        if issingle(array.dtype):
            result = empty(array.shape, dtype = np.float32)
        else:
            result = empty(array.shape, dtype = np.double)

        if array.M == 1:
            func = pu.get_angle_function(array.dtype, result.dtype, pitch = False)
            func.prepared_call(
                array._grid, array._block, result.gpudata,
                array.gpudata, array.size)
        else:
            func = pu.get_angle_function(
                array.dtype, result.dtype, pitch = True)
            func.prepared_call(
                array._grid, array._block, array.M, array.N,
                result.gpudata, result.ld, array.gpudata, array.ld)
        return result


def complex_from_amp_phase(amp, phase):
    """ Returns the angle of each element in a complex array """
    result = empty(amp.shape, dtype = floattocomplex(amp.dtype))

    if amp.M == 1:
        func = pu.get_complex_from_amp_function(amp.dtype, result.dtype, pitch = False)
        func.prepared_call(
            amp._grid, amp._block, result.gpudata,
            amp.gpudata, phase.gpudata, amp.size)
    else:
        func = pu.get_complex_from_amp_function(
            amp.dtype, result.dtype, pitch = True)
        func.prepared_call(
            amp._grid, amp._block, amp.M, amp.N,
            result.gpudata, result.ld, amp.gpudata,
            phase.gpudata, amp.ld)
    return result



import h5py
import numpy as np

import pycuda.driver as cuda
import pycuda.gpuarray as garray
from .BaseInputProcessor import BaseInputProcessor
from .ArrayInputProcessor import ArrayInputProcessor


class FileInputProcessor(ArrayInputProcessor):
    """
    Parameters
    ----------
    filename: str
              Name of the h5d file as input.
    mode:     int
              mode in BaseInputProcessor
              0: default, when the file does not cover the entire simulation,
                 inputs default to 0.
              1: when the file does not cover the entire simulation,
                 inputs defaults to the last state.
    cache_length: int
                  length (number of steps) of cache on GPU memory to preload
                  inputs.

    Examples
    --------
    Input h5d must have the following format:
    1. For 'spike_state', there are two options. First,
        '/spike_state/uids': numpy array of dtype 'S',
                             contains the uids of the nodes recorded,
                             the order of this array determines the index below
        '/spike_state/data/time': numpy array of np.float32/np.double,
                                  contains a **monotonically** increasing time stamps
                                  of spikes from all nodes
        '/spike_state/data/index': numpy array of np.int32
                                   contains the index of every entry of spike time
                                   in 'spike_state/data/time'
        The output file generated by FileOutputProcessor complies with this requirement.

       A second option is to have it as the same format as in 2 below, but have
       only float/double value of 0 (no spike at the time step) or 1 (spike)

    2. For other variables, e.g., 'I':
        '/I/uids': numpy array of dtype 'S',
                  contains the uids of the nodes recorded,
                  the order of this array determines which column each node
                  is receives as input in the 'I/data' array.
        '/I/data': numpy array of dtype np.float32/np.double,
                   contains the value of inputs injected to the nodes.
    """
    def __init__(self, filename, file_mode = 'r', mode = 0, cache_length = 1000, preload = False):
        self.filename = filename
        self.cache_length = cache_length
        inputs = {}
        self.file_mode = file_mode
        self.preload = preload
        with h5py.File(self.filename, self.file_mode) as h5file:
            var_list = []
            for var, g in h5file.items():
                if not isinstance(g, h5py.Group):
                    continue
                uids = [a.decode('utf8') for a in g.get('uids')[()].tolist()]
                if var == 'spike_state' and isinstance(g.get('data'), dict):
                    inputs[var] = {'uids': uids,
                                   'data': {'time': g.get('data/time')[:] if preload else [],
                                            'index': g.get('data/index')[:] if preload else []}}
                else:
                    inputs[var] = {'uids': uids,
                                   'data': g.get('data')[:] if preload else []}

        super(FileInputProcessor, self).__init__(inputs = inputs, mode = mode,
                                                 cache_length = cache_length)

    def add_input(self, var, uids, data):
        """
        Add inputs to the input processor

        Parameters
        ----------
        var: str
             the variable to which input will be injected to the components
        uids: list
              the uids of components to inject to
        data: ndarray or dict
              If `variable` is 'spike_state', then data can be in spike event
              format, as a dictionary {'time': np.ndarray, 'index', np.ndarray},\
              specifying the time and index of spikes.
              For all variables, data can be in stepwise input format,
              as a ndarray of shape (T, N)
              that specifies the input to each of the N components
              at every time step for a total of T steps.
              It is expected that if 'spike_state' input is added more than once,
              they must all be of one form, either in spike event format,
              or in stepwise input format.
        """
        if self.preload:
            super(FileInputProcessor, self).add_input(var, uids, data)
        else:
            if self.file_mode == 'r':
                if var not in self.variables:
                    if var == 'spike_state':
                        if isinstance(data, dict):
                            self.spike_state_format = 'event'
                        else:
                            self.spike_state_format = 'stepwise'
                else:
                    raise TypeError('Multiple {} defined in file'.format(var))
            else:
                with h5py.File(self.filename, 'r+') as h5file:
                    if var in h5file:
                        uid_length = h5file[var]['uids'].shape[0]
                        new_uids = np.concatenate((h5file[var]['uids'][:],
                                                   np.array(uids, dtype = 'S')))
                        del h5file[var]['uids']
                        h5file.create_dataset('{}/uids'.format(var),
                                              new_uids)
                        if var == 'spike_state' and isinstance(data, dict):
                            assert self.spike_state_format == 'event', \
                                   'Spike state format was previously set to stepwise in the file, must use the same format.'
                            spike_time = np.concatenate((h5file[var]['data']['time'][:],
                                                         data['time']))
                            index = np.concatenate((h5file[var]['data']['index'][:],
                                                    data['index'] + uid_length))
                            sort_order = np.argsort(spike_time)
                            del h5file[var]['data']
                            h5file.create_dataset('{}/data/index'.format(var),
                                                  maxshape = (None,),
                                                  dtype = np.int32,
                                                  data = index[sort_order])
                            h5file.create_dataset('{}/data/time'.format(var),
                                                  maxshape = (None,),
                                                  dtype = np.float64,
                                                  data = spike_time[sort_order])
                        else:
                            if var == 'spike_state':
                                assert self.spike_state_format == 'stepwise', \
                                       'Spike state format was previously set to event in the file, must use the same format.'
                            new_data = np.hstack((h5file[var]['data'][:], data))
                            del h5file[var]['data']
                            h5file.create_dataset('{}/data'.format(var),
                                                  maxshape = (None, uid_length),
                                                  dtype = np.float64,
                                                  data = new_data)
                    else:
                        uid_length = len(uids)
                        h5file.create_dataset('{}/uids'.format(var),
                                              maxshape = (None,),
                                              data = np.array(uids, dtype = 'S'))
                        if var == 'spike_state' and isinstance(data, dict):
                            self.spike_state_format = 'event'
                            h5file.create_dataset('{}/data/index'.format(var),
                                                  maxshape = (None,),
                                                  dtype = np.int32,
                                                  data = data['index'])
                            h5file.create_dataset('{}/data/time'.format(var),
                                                  maxshape = (None,),
                                                  dtype = np.float64,
                                                  data = data['time'])
                        else:
                            if var == 'spike_state':
                                self.spike_state_format = 'stepwise'
                            h5file.create_dataset('{}/data'.format(var),
                                                  dtype = np.float64,
                                                  maxshape = (None, uid_length),
                                                  data = data)
            self.add_variables([(var, uids)])

    def append_input(self, var, data):
        if preload:
            raise TypeError('FileInputProcessor in preload mode, cannot append input')
        if self.file_mode == 'r':
            raise FileIOError('File mode is set to "r", cannot write to input.')
        else:
            if var not in self.variables:
                raise KeyError('{} not in file, cannot append to. Use add_input instead.'.format(var))
            else:
                with h5py.File(self.filename, 'r+') as h5file:
                    uid_length = h5file[var]['uids'].shape[0]
                    if var == 'spike_state' and isinstance(data, dict):
                        assert self.spike_state_format == 'event', \
                               'Spike state format was previously set to stepwise in the file, must use the same format.'
                        assert data['time'][0] >= h5file[var]['data']['time'][-1], \
                               'Appended spike time must come after the spike time in spike.'
                        assert data['index'].max() < uid_length,\
                               'Spike index mismatch with uids.'
                        assert data['index'].dtype == np.int32, \
                               'index must be np.int32 type'
                        maxshape = h5file[var]['data']['time'].maxshape
                        if maxshape[0] is None:
                            h5file[var]['data']['time'].resize(
                                (h5file[var]['data']['time'].shape[0] + data['time'].shape[0]))
                            h5file[var]['data']['time'][-data['time'].shape[0]:] = data['time']
                        else:
                            spike_time = np.concatenate((h5file[var]['data']['time'][:],
                                                         data['time']))
                            del h5file[var]['data']['time']
                            h5file.create_dataset('{}/data/time'.format(var),
                                                  maxshape = (None,),
                                                  dtype = np.float64,
                                                  data = spike_time)
                        maxshape = h5file[var]['data']['time'].maxshape
                        if maxshape[0] is None:
                            h5file[var]['data']['index'].resize(
                                (h5file[var]['data']['index'].shape[0] + data['index'].shape[0]))
                            h5file[var]['data']['index'][-data['index'].shape[0]:] = data['index']
                        else:
                            index = np.concatenate((h5file[var]['data']['index'][:],
                                                    data['index']))
                            del h5file[var]['data']['index']
                            h5file.create_dataset('{}/data/index'.format(var),
                                                  maxshape = (None,),
                                                  dtype = np.int32,
                                                  data = index)
                    else:
                        if var == 'spike_state':
                            assert self.spike_state_format == 'stepwise', \
                                   'Spike state format was previously set to event in the file, must use the same format.'
                        assert data.shape[1] == uid_length, \
                               'Data dimension mismatch with uids.'
                        assert data.dtype == np.double, 'data must be of np.double'
                        maxshape = h5file[var]['data'].maxshape
                        if maxshape[0] is None:
                            h5file[var]['data'].resize(
                                (h5file[var]['data'].shape[0] + data.shape[0],
                                 uid_length))
                            h5file[var]['data'][-data.shape[0]:,:] = data
                        else:
                            new_data = np.vstack((h5file[var]['data'][:],
                                                  data))
                            del h5file[var]['data']
                            h5file.create_dataset('{}/data'.format(var),
                                                  dtype = np.float64,
                                                  maxshape = (None, uid_length),
                                                  data = new_data)

    def pre_run(self):
        if self.preload:
            super(FileInputProcessor, self).pre_run()
        else:
            self.h5file = h5py.File(self.filename, 'r')
            self.dsets = {}
            self.cache = {}
            self.counts = {}
            self.end_of_var_in_array = {}
            self.block_total = {}
            self.end_of_var = {}
            self.last_read_index = {}
            for var, g in self.h5file.items():
                if not isinstance(g, h5py.Group):
                    continue
                self.block_total[var] = 0
                if var == 'spike_state':
                    if 'index' in g.get('data'):
                        self.dsets[var] = {'index': g.get('data/index')[:] if self.preload else g.get('data/index'),
                                           'time': g.get('data/time')[:] if self.preload else g.get('data/time')}
                        self.spike_state_format = 'event'
                    else:
                        self.dsets[var] = g.get('data')[:] if self.preload else g.get('data')
                        self.spike_state_format = 'stepwise'
                    self.cache[var] = garray.empty((self.cache_length, len(g['uids'])),
                                                   self.variables[var]['input'].dtype)
                else:
                    self.dsets[var] = g.get('data')[:] if self.preload else g.get('data')
                    self.cache[var] = garray.empty((self.cache_length, len(g['uids'])),
                                                   self.variables[var]['input'].dtype)
                self.last_read_index[var] = 0
                self.counts[var] = 0
                self.end_of_var[var] = False
                self.end_of_var_in_array[var] = False

    def update_input(self):
        super(FileInputProcessor, self).update_input()
        if not self.preload:
            if self.end_of_input:
                self.h5file.close()

    def post_run(self):
        if not self.preload:
            if not self.end_of_input:
                self.h5file.close()




from future.utils import iteritems

from .utils import parray
import pycuda.gpuarray as garray
from pycuda.tools import dtype_to_ctype, context_dependent_memoize
from pycuda.compiler import SourceModule
import pycuda.driver as cuda
# import pycuda.elementwise as elementwise

import numpy as np
import numbers

class MemoryManager(object):
    def __init__(self,devices=None):
        '''
        TODO : support multiple devices feature. This probably will require
        changes to the neurokernel core as well
        devices should be a list containing the device numbers of the GPUs to be
        used by this MemoryManager
        '''
        self.devices = devices
        self.variables = {}
        self.parameters = {}
        self.mapping = {}          #Mapping from [model_name->variable/parameter]->pos

    def get_buffer(self, variable_name):
        return self.variables[variable_name]['buffer']

    def mutate_variable(self, variable_name, transform):
        pass

    def fill_zeros(self, variable=None, model=None):
        # TODO: Cache dest_inds based on variable and model
        assert(variable or model)
        if variable and not model:
            assert(variable in self.variables)
            d = self.variables[variable]
            dest_inds = np.arange(0, d['cumlen'][-1],1).astype(np.int32)
            buff = d['buffer']
            dest_mem = garray.GPUArray((1,d['cumlen'][-1]),buff.dtype,
                                       gpudata=int(buff.gpudata)+\
                                       buff.current*buff.ld*\
                                       buff.dtype.itemsize)
            dest_mem.fill(0)
        elif model and not variable:
            for var, d in iteritems(self.variables):
                if model in d['models']:
                    mind = d['models'].index(model)
                    stind = d['cumlen'][mind]
                    dest_inds = np.arange(stind, stind+d['len'][mind],1).astype(np.int32)
                    buff = d['buffer']
                    # dest_mem = garray.GPUArray((1,buff.size),buff.dtype,
                    #                            gpudata=int(buff.gpudata)+\
                    #                            buff.current*buff.ld*\
                    #                            buff.dtype.itemsize)
                    dest_mem = int(buff.gpudata)+buff.current*buff.ld*buff.dtype.itemsize
                    self._fill_zeros_kernel(var, model, dest_mem)
        else:
            assert(variable in self.variables)
            d = self.variables[variable]
            if model in d['models']:
                mind = d['models'].index(model)
                stind = d['cumlen'][mind]
                dest_inds = np.arange(stind, stind+d['len'][mind],1).astype(np.int32)
                buff = d['buffer']
                # dest_mem = garray.GPUArray((1,buff.size),buff.dtype,
                #                            gpudata=int(buff.gpudata)+\
                #                            buff.current*buff.ld*\
                #                            buff.dtype.itemsize)
                dest_mem = int(buff.gpudata)+buff.current*buff.ld*buff.dtype.itemsize
                self._fill_zeros_kernel(variable, model, dest_mem)

    def precompile_fill_zeros(self):
        self.fill_zeros_func = {}
        self.dest_inds = {}
        for var, d in iteritems(self.variables):
            self.fill_zeros_func[var] = get_fill_zeros_kernel(d['buffer'].dtype)
            self.dest_inds[var] = {}
            for model in d['models']:
                mind = d['models'].index(model)
                stind = d['cumlen'][mind]
                dest_inds = np.arange(stind, stind+d['len'][mind],1).astype(np.int32)
                self.dest_inds[var][model] = garray.to_gpu(dest_inds)

    def mutate_parameter(self, model_name, param, transform):
        pass

    def memory_alloc(self, variable_name, size, buffer_length=1,
                     dtype=np.double, info={}, init=None):
        assert(variable_name not in self.variables)
        self.variables[variable_name] = {'buffer': \
                            CircularArray(size, buffer_length, dtype, init)}
        self.variables[variable_name].update(info)

    def params_htod(self, model_name, param_dict, dtype=np.double):
        if model_name in self.parameters:
            assert(not (set(self.parameters[model_name]) &
                        set(param_dict)))
        else:
            self.parameters[model_name] = {}

        for k, v in param_dict.items():
            if k in ['pre','npre','cumpre']:
                self.parameters[model_name][k] = \
                                {var: garray.to_gpu(np.array(v[var],np.int32))\
                                 for var in v}
                continue
            if k=='conn_data':
                cd = {}
                for var,data in v.items():
                    cd[var] = {}
                    for d_key,d in data.items():
                        if isinstance(d, list):
                            tmp = np.array(d)
                            if not np.issubdtype(tmp.dtype, np.number):
                                continue
                        elif isinstance(d, dict):
                            continue
                        else:
                            continue
                        if d_key=='delay':
                            cd[var][d_key] = garray.to_gpu(tmp.astype(np.int32))
                        else:
                            cd[var][d_key] = garray.to_gpu(tmp.astype(dtype))
                self.parameters[model_name]['conn_data'] = cd
            if isinstance(v, list):
                tmp = np.array(v)
                if not np.issubdtype(tmp.dtype, np.number):
                    continue
            elif isinstance(v, dict):
                continue
            else:
                continue
            self.parameters[model_name][k] = garray.to_gpu(tmp.astype(dtype))

    def step(self):
        for d in self.variables.values():
            d['buffer'].step()

    def _fill_zeros_kernel(self, var, model, dest):
        func = self.fill_zeros_func[var]
        inds = self.dest_inds[var][model]
        func.prepared_async_call(
            func.grid, func.block, None,
            dest, inds.gpudata, inds.size)

    # def _fill_zeros_kernel(self, dest, inds):
    #     """
    #     Set `dest[inds[i]] = 0 for i in range(len(inds))`
    #     """
    #
    #     try:
    #         func = self._fill_zeros_kernel.cache[(inds.dtype, dest.dtype)]
    #     except KeyError:
    #         inds_ctype = dtype_to_ctype(inds.dtype)
    #         data_ctype = dtype_to_ctype(dest.dtype)
    #         v = ("{data_ctype} *dest," +\
    #              "{inds_ctype} *inds").format(\
    #                     data_ctype=data_ctype,inds_ctype=inds_ctype)
    #         func = elementwise.ElementwiseKernel(v,\
    #         "dest[inds[i]] =0")
    #         self._fill_zeros_kernel.cache[(inds.dtype, dest.dtype)] = func
    #     func(dest, inds, range=slice(0, len(inds), 1) )
    #
    # _fill_zeros_kernel.cache = {}


@context_dependent_memoize
def get_fill_zeros_kernel(data_dtype):
    template = """
__global__ void update(%(data_ctype)s* dest,
                       %(inds_ctype)s* inds,
                       int N)
{
    int tid = threadIdx.x + blockIdx.x * blockDim.x;
    int total_threads = gridDim.x * blockDim.x;

    for(int i = tid; i < N; i += total_threads)
    {
        dest[inds[i]] = 0;
    }
}
"""
    mod = SourceModule(template % {"data_ctype": dtype_to_ctype(data_dtype),
                                   "inds_ctype": dtype_to_ctype(np.int32)})
    func = mod.get_function("update")
    func.prepare('PPi')
    func.block = (128,1,1)
    func.grid = (16 * cuda.Context.get_device().MULTIPROCESSOR_COUNT, 1)
    return func


class CircularArray(object):
    """
    Circular buffer to support variables with memory

    Parameters
    ----------
    size : int
        size of the array for each step
    buffer_length : int
        Number of steps into the past to buffer
    dtype : np.dtype
        Data type to be used for the array
    init : dtype
        Initial value for the data. If not specified defaults to zero
    Attributes
    ----------
    size : int
        See above
    buffer_length : int
        See above
    dtype :
        See above
    parr : parray
        Pitched array of dimensions (buffer_length, size)
    current : int
        An integer in [0,buffer_length) representing Current position
        in the buffer.
    Methods
    -------
    step()
        Advance indices of current position in the buffer
    """

    def __init__(self, size, buffer_length, dtype=np.double, init=None):

        self.size = size
        if not isinstance(dtype, np.dtype): dtype = np.dtype(dtype)
        self.dtype = dtype

        self.buffer_length = buffer_length
        if init:
            try:
                init = dtype(init)
                self.parr = parray.ones(
                 (buffer_length, size), dtype) * init
            except:
                self.parr = parray.zeros(
                 (buffer_length, size), dtype)
        else:
            self.parr = parray.zeros(
                 (buffer_length, size), dtype)
        self.current = 0
        self.gpudata = self.parr.gpudata
        self.ld = self.parr.ld


    def step(self):
        """
        Advance indices of current graded potential and spiking neuron values.
        """

        if self.size > 0:
            self.current += 1
            if self.current >= self.buffer_length:
                self.current = 0




from collections import OrderedDict

import numpy as np

from neurokernel.LPU.NDComponents.MembraneModels.BaseMembraneModel import BaseMembraneModel

class MorrisLecar(BaseMembraneModel):
    params = ['V1', 'V2', 'V3', 'V4', 'phi', 'offset',
              'V_L', 'V_Ca', 'V_K', 'g_L', 'g_Ca', 'g_K']
    extra_params = ['initV', 'initn']
    internals = OrderedDict([('V', -70.0), ('n', 0.3525)])

    @property
    def maximum_dt_allowed(self):
        return 1e-5

    def pre_run(self, update_pointers):
        super(MorrisLecar, self).pre_run(update_pointers)
        self.add_initializer('initn', 'n', update_pointers)

    def get_update_template(self):
        template = """
__device__ %(internal_n)s compute_n(%(update_V)s V, %(internal_n)s n, %(param_V3)s V3, %(param_V4)s V4, %(param_phi)s phi)
{
    %(internal_n)s n_inf = 0.5 * (1 + tanh((V - V3) / V4));
    %(internal_n)s dn = phi * cosh(( V - V3) / (V4*2)) * (n_inf - n);
    return dn;
}

__device__ %(update_V)s compute_V(%(update_V)s V, %(internal_n)s n, %(input_I)s I, %(param_V1)s V1, %(param_V2)s V2,
                           %(param_offset)s offset, %(param_V_L)s V_L, %(param_V_Ca)s V_Ca,
                           %(param_V_K)s V_K, %(param_g_L)s g_L, %(param_g_K)s g_K, %(param_g_Ca)s g_Ca)
{
    %(update_V)s m_inf = 0.5 * (1+tanh((V - V1)/V2));
    %(update_V)s dV = (I - g_L * (V - V_L) - g_K * n * (V - V_K) - g_Ca * m_inf * (V - V_Ca) + offset);
    return dV;
}

__global__ void
update(int num_comps, %(dt)s dt, int nsteps,
                      %(input_I)s* g_I, %(param_V1)s* g_V1, %(param_V2)s* g_V2, %(param_V3)s* g_V3,
                      %(param_V4)s* g_V4, %(param_phi)s* g_phi, %(param_offset)s* g_offset,
                      %(param_V_L)s* g_V_L, %(param_V_Ca)s* g_V_Ca, %(param_V_K)s* g_V_K,
                      %(param_g_L)s* g_g_L, %(param_g_Ca)s* g_g_Ca, %(param_g_K)s* g_g_K,
                      %(internal_V)s* g_internalV, %(internal_n)s* g_n, %(update_V)s* g_V)
{
    int tid = threadIdx.x + blockIdx.x * blockDim.x;
    int total_threads = gridDim.x * blockDim.x;

    %(update_V)s V, dV;
    %(internal_n)s n, dn;
    %(input_I)s I;
    %(param_V1)s V1;
    %(param_V2)s V2;
    %(param_V3)s V3;
    %(param_V4)s V4;
    %(param_phi)s phi;
    %(param_offset)s offset;
    %(param_V_L)s V_L;
    %(param_V_Ca)s V_Ca;
    %(param_V_K)s V_K;
    %(param_g_L)s g_L;
    %(param_g_Ca)s g_Ca;
    %(param_g_K)s g_K;

    for(int k = tid; k < num_comps; k += total_threads)
    {
        V = g_internalV[k];
        n = g_n[k];
        V1 = g_V1[k];
        V2 = g_V2[k];
        V3 = g_V3[k];
        V4 = g_V4[k];
        phi = g_phi[k];
        offset = g_offset[k];
        V_L = g_V_L[k];
        V_Ca = g_V_Ca[k];
        V_K = g_V_K[k];
        g_L = g_g_L[k];
        g_Ca = g_g_Ca[k];
        g_K = g_g_K[k];
        I = g_I[k];

        for(int i = 0; i < nsteps; ++i)
        {
            dn = compute_n(V, n, V3, V4, phi);
            dV = compute_V(V, n, I, V1, V2,
                           offset, V_L, V_Ca, V_K,
                           g_L, g_K, g_Ca);
            V += dV * dt;
            n += dn * dt;
        }

        g_V[k] = V;
        g_internalV[k] = V;
        g_n[k] = n;
    }
}
"""
        return template


if __name__ == '__main__':
    import argparse
    import itertools
    import networkx as nx
    from neurokernel.tools.logging import setup_logger
    import neurokernel.core_gpu as core

    from neurokernel.LPU.LPU import LPU

    from neurokernel.LPU.InputProcessors.StepInputProcessor import StepInputProcessor
    from neurokernel.LPU.OutputProcessors.FileOutputProcessor import FileOutputProcessor

    import neurokernel.mpi_relaunch

    dt = 1e-4
    dur = 1.0
    steps = int(dur/dt)

    parser = argparse.ArgumentParser()
    parser.add_argument('--debug', default=False,
                        dest='debug', action='store_true',
                        help='Write connectivity structures and inter-LPU routed data in debug folder')
    parser.add_argument('-l', '--log', default='none', type=str,
                        help='Log output to screen [file, screen, both, or none; default:none]')
    parser.add_argument('-s', '--steps', default=steps, type=int,
                        help='Number of steps [default: %s]' % steps)
    parser.add_argument('-g', '--gpu_dev', default=0, type=int,
                        help='GPU device number [default: 0]')
    args = parser.parse_args()

    file_name = None
    screen = False
    if args.log.lower() in ['file', 'both']:
        file_name = 'neurokernel.log'
    if args.log.lower() in ['screen', 'both']:
        screen = True
    logger = setup_logger(file_name=file_name, screen=screen)

    man = core.Manager()

    G = nx.MultiDiGraph()

    G.add_node('neuron0', **{
               'class': 'MorrisLecar',
               'name': 'MorrisLecar',
               'V1': -20.,
               'V2': 50.,
               'V3': -40.,
               'V4': 20.,
               'phi': 0.001,
               'offset': 0.,
               'V_L': -40.,
               'V_Ca': 120.,
               'V_K': -80.,
               'g_L': 3.,
               'g_Ca': 4.,
               'g_K': 16.,
               'initV': -46.080,
               'initn': 0.3525
               })

    comp_dict, conns = LPU.graph_to_dicts(G)

    fl_input_processor = StepInputProcessor('I', ['neuron0'], 5, 0.2, 0.4)
    fl_output_processor = FileOutputProcessor([('V', None)], 'new_output.h5', sample_interval=1)

    man.add(LPU, 'ge', dt, comp_dict, conns,
            device=args.gpu_dev, input_processors = [fl_input_processor],
            output_processors = [fl_output_processor], debug=args.debug)

    man.spawn()
    man.start(steps=args.steps)
    man.wait()



#!/usr/bin/env python

import sys, os
from glob import glob

# Install setuptools if it isn't available:
try:
    import setuptools
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()

from distutils.command.install_headers import install_headers
from setuptools import find_packages
from setuptools import setup

NAME =               'neurodriver'
VERSION =            '0.3.0'
AUTHOR =             'Neurokernel Development Team'
AUTHOR_EMAIL =       'neurokernel-dev@columbia.edu'
URL =                'https://github.com/neurokernel/neurodriver/'
MAINTAINER =         AUTHOR
MAINTAINER_EMAIL =   AUTHOR_EMAIL
DESCRIPTION =        'LPU class with support for various neuron and synapse models'
LONG_DESCRIPTION =   DESCRIPTION
DOWNLOAD_URL =       URL
LICENSE =            'BSD'
CLASSIFIERS = [
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: BSD License',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Topic :: Scientific/Engineering',
    'Topic :: Software Development']

# Explicitly switch to parent directory of setup.py in case it
# is run from elsewhere:
os.chdir(os.path.dirname(os.path.realpath(__file__)))
PACKAGES =           find_packages()

if __name__ == "__main__":
    if os.path.exists('MANIFEST'):
        os.remove('MANIFEST')

    setup(
        name = NAME,
        version = VERSION,
        author = AUTHOR,
        author_email = AUTHOR_EMAIL,
        license = LICENSE,
        classifiers = CLASSIFIERS,
        description = DESCRIPTION,
        long_description = LONG_DESCRIPTION,
        url = URL,
        maintainer = MAINTAINER,
        maintainer_email = MAINTAINER_EMAIL,
        packages = PACKAGES,
        include_package_data = True,
        install_requires = [
            'neurokernel >= 0.3.0',
        ],
        )




from collections import OrderedDict
import numpy as np
import h5py
from datetime import datetime
from neurokernel.LPU.OutputProcessors.BaseOutputProcessor import BaseOutputProcessor
import pycuda.driver as cuda
import pycuda.gpuarray as garray

class OutputRecorder(BaseOutputProcessor):
    """
    Store specified output in memory rather than writing to file.

    Parameters
    ----------
    var_list: list
    sample_interval: int
    cache_length: int

    Examples
    --------
    import networx as np
    from neurokernel.LPU.InputProcessors.StepInputProcessor import StepInputProcessor
    from neurokernel.LPU.OutputProcessors.OutputRecorder import OutputRecorder

    dt = 1e-6
    dur = 1.0
    steps = int(dur/dt)

    G = nx.MultiDiGraph()
    G.add_node('neuron0', model = 'HodgkinHuxley2',
               g_K = 36.0, g_Na = 120.0, g_L = 0.3,
               E_Na = 50.0, E_K = -77.0, E_L = -54.354)
    input_processor = StepInputProcessor('I', ['neuron0'], 20.0, 0.2, 0.8)
    output_processor = OutputRecorder(
                        [('V', ['neuron0']), ('spike_state', ['neuron0'])],
                        sample_interval = 100) # sample at 1/(dt*100) = 1e4 sampling rate

    lpu = LPU(dt, 'obj', {'graph': G, 'kwargs': {'class_key': 'model'}},
              device = 0, id = 'test', input_processors=[input_processor],
              output_processors = [output_processor],
              manager = False)
    lpu.run(steps = steps)

    neuron0_response = output_processor.get_output(uids = ['neuron0'])
    print(neuron0_response['neuron0'])

    voltage_response = output_processor.get_output(var = 'V')
    print(voltage_response)

    spike_times = output_processor.get_output(var = 'spike_state', uids = 'neuron0')
    print(spike_times['neuron0'])

    output_processor.to_file('test.h5')
    """
    def __init__(self, var_list, sample_interval = 1, cache_length = 1000):
        super(OutputRecorder, self).__init__(var_list, sample_interval)
        self._output = {}
        self._cache = {}
        self.cache_length = 1000

    @property
    def output(self):
        return self._output

    def pre_run(self):
        for var, d in self.variables.items():
            if var == 'spike_state':
                self.output[var] = {'data':
                                        {'time': [],
                                         'index': []},
                                    'uids': d['uids']}
            else:
                self.output[var] = {'data': [],
                                    'uids': d['uids']}
            self._cache[var] = garray.empty((self.cache_length, len(d['uids'])),
                                           d['output'].dtype)
        self.count = 0
        self.scount = 0
        self.stime_shift = 0

    def get_output_array(self, var):
        if var == 'spike_state':
            return int(self._cache[var].gpudata)+\
                    self.scount*self._cache[var].shape[1]*self._cache[var].dtype.itemsize
        else:
            return int(self._cache[var].gpudata)+\
                   self.count*self._cache[var].shape[1]*self._cache[var].dtype.itemsize

    def process_output(self):
        self.count += 1
        if self.count == self.cache_length:
            for var, d in self.variables.items():
                if var != 'spike_state':
                    self.output[var]['data'].append(self._cache[var].get())
            self.count = 0

    def process_spike_output(self):
        self.scount += 1
        if self.scount == self.cache_length:
            var = 'spike_state'
            t, index = np.where(self._cache[var].get()!=0)
            t = t*self.sim_dt
            nspikes = index.size
            if nspikes:
                self.output[var]['data']['time'].append(t+self.stime_shift)
                self.output[var]['data']['index'].append(index.astype(np.int32))
            self.stime_shift += self.cache_length*self.sim_dt
            self.scount = 0

    def post_run(self):
        if self.count > 0:
            for var, d in self.variables.items():
                if var != 'spike_state':
                    self.output[var]['data'].append(self._cache[var].get()[:self.count,:])
        if self.scount > 0:
            var = 'spike_state'
            t, index = np.where(self._cache[var].get()[:self.scount,:]!=0)
            t = t * self.sim_dt
            nspikes = index.size
            if nspikes:
                self.output[var]['data']['time'].append(t+self.stime_shift)
                self.output[var]['data']['index'].append(index.astype(np.int32))
        for var, d in self.variables.items():
            if var != 'spike_state':
                self.output[var]['data'] = np.vstack(self.output[var]['data'])
            else:
                if len(self.output[var]['data']['time']):
                    self.output[var]['data']['time'] = np.concatenate(self.output[var]['data']['time'])
                    self.output[var]['data']['index'] = np.concatenate(self.output[var]['data']['index'])
                else:
                    self.output[var]['data']['time'] = np.zeros(0, np.double)
                    self.output[var]['data']['index'] = np.zeros(0, np.int32)

    def get_output(self, var = None, uids = None):
        """
        retrieve outputs by variable name and uids

        Parameters
        ----------
        var: str
             Name of the variable to retrieve.
             If None, all variables associated with a uid will be retrieved.
        uids: str or list of str
              uids of the component to retrieve
              If not specified (None), all uids associated with the `var`
              will be retrieved.

        Returns
        -------
        output: dict or OrderedDict
                If uids is a list or tuple, returns a OrderedDict, Otherwise a dict,
                with output keyed by uid, and values are either the data/spike_times
                of the component, or a dict keyed by variable name and data/spike_times
                in the value.
        """
        if var is None and uids is None:
            return self.output
        elif var is not None and uids is None:
            return self._get_output_by_var(var)
        elif var is None and uids is not None:
            return self._get_output_by_uids(uids)
        else:
            return self._get_output_by_var_and_uids(var, uids)

    def _get_output_by_var(self, var):
        uids = self.output[var]['uids']
        data = self.output[var]['data']
        if var == 'spike_state':
            output = {uid: {'data': data['time'][data['index']==i] + self.start_time} for i, uid in enumerate(uids)}
        else:
            t = np.arange(0, data.shape[0])*self.dt*self.sample_interval + self.start_time
            output = {uid: {'time': t, 'data': data[:,i].copy()} for i, uid in enumerate(uids)}
        return output

    def _get_output_by_uid(self, uid):
        output = {}
        for var in self.output:
            try:
                index = self.output[var]['uids'].index(uid)
            except ValueError:
                pass
            else:
                if var == 'spike_state':
                    output[var] = {'data': self.output[var]['data']['time'][self.output[var]['data']['index'] == index] + self.start_time}
                else:
                    t = np.arange(0, self.output[var]['data'].shape[0])*self.dt*self.sample_interval + self.start_time
                    output[var] = {'time': t, 'data': self.output[var]['data'][:,index].copy()}
        return output

    def _get_output_by_uids(self, uids):
        if isinstance(uids, str):
            return {uids: self._get_output_by_uid(uids)}
        elif isinstance(uids, (list, tuple)):
            return OrderedDict([(uid, self._get_output_by_uid(uid)) for uid in uids])
        elif isinstance(uids, set):
            return {uid: self._get_output_by_uid(uid) for uid in uids}

    def _get_output_by_var_and_uid(self, var, uid):
        output = None
        try:
            index = self.output[var]['uids'].index(uid)
        except ValueError:
            pass
        else:
            if var == 'spike_state':
                output = {'data': self.output[var]['data']['time'][self.output[var]['data']['index'] == index] + self.start_time}
            else:
                t = np.arange(0, self.output[var]['data'].shape[0])*self.dt*self.sample_interval + self.start_time
                output = {'time': t, 'data': self.output[var]['data'][:,index].copy()}
        return output

    def _get_output_by_var_and_uids(self, var, uids):
        if isinstance(uids, str):
            return {uids: self._get_output_by_var_and_uid(var, uids)}
        elif isinstance(uids, (list, tuple)):
            return OrderedDict([(uid, self._get_output_by_var_and_uid(var, uid)) for uid in uids])
        elif isinstance(uids, set):
            return {uid: self._get_output_by_var_and_uid(var, uid) for uid in uids}

    def to_file(self, filename):
        """
        Write recorded outputs to file with name `filename`,
        compatible with the file written directly by
        `.FileOutputProcessor`

        parameters
        ----------
        filename: str
                  name of the file to store the results
        """
        with h5py.File(filename, 'w') as h5file:
            h5file.create_dataset('metadata', (), 'i')
            h5file['metadata'].attrs['start_time'] = self.start_time
            h5file['metadata'].attrs['sample_interval'] = self.sample_interval
            h5file['metadata'].attrs['dt'] = self.dt
            h5file['metadata'].attrs[
            'DateCreated'] = datetime.now().isoformat()

            for var, d in self.variables.items():
                if var == 'spike_state':
                    h5file.create_dataset(var + '/data/index',
                                          dtype = np.int32,
                                          maxshape = (None,),
                                          data = self.output[var]['data']['index'])
                    h5file.create_dataset(var + '/data/time',
                                          dtype = np.double,
                                          maxshape = (None,),
                                          data = self.output[var]['data']['time'])
                else:
                    h5file.create_dataset(var + '/data',
                                               dtype = self.output[var]['data'].dtype,
                                               maxshape=(None, len(d['uids'])),
                                               data = self.output[var]['data'])
                h5file.create_dataset(var + '/uids',
                                      data = np.array(d['uids'], dtype = 'S'))
